@@ -1,6 +1,6 @@
 package cPanel::PublicAPI;
 
-# Copyright (c) 2015, cPanel, Inc.
+# Copyright 2017, cPanel, Inc.
 # All rights reserved.
 # http://cpanel.net
 #
@@ -29,15 +29,14 @@ package cPanel::PublicAPI;
 # OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-our $VERSION = 1.3;
+our $VERSION = '2.2.1';
 
 use strict;
-use Socket           ();
-use Carp             ();
-use MIME::Base64     ();
-use IO::Socket::INET ();
+use Carp            ();
+use MIME::Base64    ();
+use HTTP::Tiny      ();
+use HTTP::CookieJar ();
 
-my ( $whm_sock, $whm_sock_host, $loaded_ssl );
 our %CFG;
 
 my %PORT_DB = (
@@ -63,25 +62,7 @@ sub new {
 
     $self->{'debug'}   = $OPTS{'debug'}   || 0;
     $self->{'timeout'} = $OPTS{'timeout'} || 300;
-
-    if ( exists $OPTS{'usessl'} ) {
-        $self->{'usessl'} = $OPTS{'usessl'};
-    }
-    else {
-        $self->{'usessl'} = 1;
-    }
-    if ( exists $OPTS{'ssl_verify_mode'} ) {
-        $self->{'ssl_verify_mode'} = $OPTS{'ssl_verify_mode'};
-    }
-    else {
-        $self->{'ssl_verify_mode'} = 1;
-    }
-    if ( exists $OPTS{'keepalive'} ) {
-        $self->{'keepalive'} = int $OPTS{'keepalive'};
-    }
-    else {
-        $self->{'keepalive'} = 0;
-    }
+    $self->{'usessl'} = exists $OPTS{'usessl'} ? $OPTS{'usessl'} : 1;
 
     if ( exists $OPTS{'ip'} ) {
         $self->{'ip'} = $OPTS{'ip'};
@@ -93,6 +74,13 @@ sub new {
         $self->{'ip'} = '127.0.0.1';
     }
 
+    $self->{'ua'} = HTTP::Tiny->new(
+        agent      => "cPanel::PublicAPI/$VERSION ",
+        verify_SSL => ( exists $OPTS{'ssl_verify_mode'} ? $OPTS{'ssl_verify_mode'} : 1 ),
+        keep_alive => ( exists $OPTS{'keepalive'} ? int $OPTS{'keepalive'} : 0 ),
+        timeout    => $self->{'timeout'},
+    );
+
     if ( exists $OPTS{'error_log'} && $OPTS{'error_log'} ne 'STDERR' ) {
         if ( !open( $self->{'error_fh'}, '>>', $OPTS{'error_log'} ) ) {
             print STDERR "Unable to open $OPTS{'error_log'} for writing, defaulting to STDERR for error logging: $@\n";
@@ -112,6 +100,15 @@ sub new {
         $self->debug("Setting user based on current uid ($>)") if $self->{'debug'};
     }
 
+    if ( exists $OPTS{'api_token'} && exists $OPTS{'accesshash'} ) {
+        $self->error('You cannot specify both an accesshash and an API token');
+        die $self->{'error'};
+    }
+
+    # Allow the user to specify an api_token instead of an accesshash.
+    # Though, it will just act as a synonym.
+    $OPTS{'accesshash'} = $OPTS{'api_token'} if $OPTS{'api_token'};
+
     if ( ( !exists( $OPTS{'pass'} ) || $OPTS{'pass'} eq '' ) && ( !exists $OPTS{'accesshash'} || $OPTS{'accesshash'} eq '' ) ) {
         my $homedir = exists $INC{'Cpanel/PwCache.pm'} ? ( Cpanel::PwCache::getpwuid($>) )[7] : ( getpwuid($>) )[7];
         $self->debug("Attempting to detect correct authentication credentials") if $self->{'debug'};
@@ -122,19 +119,18 @@ sub new {
                 $self->{'accesshash'} = readline($hash_fh);
                 $self->{'accesshash'} =~ s/[\r\n]+//g;
                 close($hash_fh);
-                $self->debug("Got access hash from $homedir/.accesshash") if $self->{'debug'};
+                $self->debug("Got accesshash from $homedir/.accesshash") if $self->{'debug'};
             }
             else {
-                $self->debug("Failed to fetch access hash from $homedir/.accesshash") if $self->{'debug'};
+                $self->debug("Failed to fetch accesshash from $homedir/.accesshash") if $self->{'debug'};
             }
         }
         elsif ( exists $ENV{'REMOTE_PASSWORD'} && $ENV{'REMOTE_PASSWORD'} && $ENV{'REMOTE_PASSWORD'} ne '__HIDDEN__' && exists $ENV{'SERVER_SOFTWARE'} && $ENV{'SERVER_SOFTWARE'} =~ /^cpsrvd/ ) {
             $self->debug("Got user password from the REMOTE_PASSWORD environment variables.") if $self->{'debug'};
-            $OPTS{'pass'} = $ENV{'REMOTE_PASSWORD'};
+            $self->{'pass'} = $ENV{'REMOTE_PASSWORD'};
         }
-
         else {
-            Carp::confess('pass or access hash is a required parameter');
+            Carp::confess('pass, accesshash, or api_token is a required parameter');
         }
     }
     elsif ( $OPTS{'pass'} ) {
@@ -147,6 +143,8 @@ sub new {
         $self->debug("Using accesshash param from object creation") if $self->{'debug'};
     }
 
+    $self->_update_operating_mode();
+
     return $self;
 }
 
@@ -164,12 +162,18 @@ sub pass {
     my $self = shift;
     $self->{'pass'} = shift;
     delete $self->{'accesshash'};
+    $self->_update_operating_mode();
 }
 
 sub accesshash {
     my $self = shift;
     $self->{'accesshash'} = shift;
     delete $self->{'pass'};
+    $self->_update_operating_mode();
+}
+
+sub api_token {
+    return shift->accesshash(@_);
 }
 
 sub whm_api {
@@ -181,6 +185,15 @@ sub whm_api {
     if ( defined $format && $format ne 'xml' && $format ne 'json' && $format ne 'ref' ) {
         $self->error("cPanel::PublicAPI::whm_api_request() was called with an invalid data format, the only valid format are 'json', 'ref' or 'xml'");
     }
+
+    $formdata ||= {};
+    if ( ref $formdata ) {
+        $formdata = { 'api.version' => 1, %$formdata };
+    }
+    elsif ( $formdata !~ /(^|&)api\.version=/ ) {
+        $formdata = "api.version=1&$formdata";
+    }
+
     my $query_format;
     if ( defined $format ) {
         $query_format = $format;
@@ -192,233 +205,126 @@ sub whm_api {
     my $uri = "/$query_format-api/$call";
 
     my ( $status, $statusmsg, $data ) = $self->api_request( 'whostmgr', $uri, 'POST', $formdata );
-    if ( $self->{'error'} ) {
-        die $self->{'error'};
-    }
-    if ( defined $format && ( $format eq 'json' || $format eq 'xml' ) ) {
-        return $$data;
-    }
-    else {
-        my $parsed_data;
-        eval { $parsed_data = $CFG{'serializer_can_deref'} ? $CFG{'api_serializer_obj'}->($data) : $CFG{'api_serializer_obj'}->($$data); };
-        if ( !ref $parsed_data ) {
-            $self->error("There was an issue with parsing the following response from cPanel or WHM: [data=[$$data]]");
-            die $self->{'error'};
+    return $self->_parse_returndata(
+        {
+            'caller' => 'whm_api',
+            'data'   => $data,
+            'format' => $format,
+            'call'   => $call
         }
-        if (
-            ( exists $parsed_data->{'error'} && $parsed_data->{'error'} =~ /Unknown App Requested/ ) ||    # xml-api v0 version
-            ( exists $parsed_data->{'metadata'}->{'reason'} && $parsed_data->{'metadata'}->{'reason'} =~ /Unknown app requested/ )
-          ) {                                                                                              # xml-api v1 version
-            $self->error("cPanel::PublicAPI::whm_api was called with the invalid API call of: $call.");
-            return;
-        }
-        return $parsed_data;
-    }
+    );
 }
 
 sub api_request {
     my ( $self, $service, $uri, $method, $formdata, $headers ) = @_;
 
-    $self->{'error'} = 0;
-
     $formdata ||= '';
     $method   ||= 'GET';
-    $headers  ||= '';
+    $headers  ||= {};
 
     $self->debug("api_request: ( $self, $service, $uri, $method, $formdata, $headers )") if $self->{'debug'};
 
     $self->_init() if !exists $CFG{'init'};
 
-    $self->{'error'} = undef;
+    undef $self->{'error'};
     my $timeout = $self->{'timeout'} || 300;
 
-    my $authstr;
-    if ( exists $self->{'accesshash'} ) {
-        $self->{'accesshash'} =~ s/[\r\n]//g;
-        $authstr = 'WHM ' . $self->{'user'} . ':' . $self->{'accesshash'};
-    }
-    elsif ( exists $self->{'pass'} ) {
-        $authstr = 'Basic ' . MIME::Base64::encode_base64( $self->{'user'} . ':' . $self->{'pass'} );
-        $authstr =~ s/[\r\n]//g;
-    }
-    else {
-        die 'No user name or password set, cannot execute query.';
-    }
-
     my $orig_alarm = 0;
     my $page;
-    my $port;
-
-    if ( $self->{'usessl'} ) {
-        if ( !$loaded_ssl ) {
-            eval { require IO::Socket::SSL; };
-            if ($@) {
-                die "Failed to load IO::Socket::SSL: $@.";
-            }
-            Net::SSLeay::load_error_strings();
-            Net::SSLeay::OpenSSL_add_ssl_algorithms();
-            Net::SSLeay::ERR_load_crypto_strings();
-            Net::SSLeay::randomize();
-            IO::Socket::SSL->import('inet4');    # see case 16674
-            $loaded_ssl = 1;
-        }
-        IO::Socket::SSL->import('inet4');        # see case 16674
-        $port = $service =~ /^\d+$/ ? $service : $PORT_DB{$service}{'ssl'};
-    }
-    else {
-        $port = $service =~ /^\d+$/ ? $service : $PORT_DB{$service}{'plaintext'};
-    }
 
+    my $port = $self->_determine_port_for_service($service);
     $self->debug("Found port for service $service to be $port (usessl=$self->{'usessl'})") if $self->{'debug'};
 
     eval {
-        if ( !$self->{'user'} ) {
-            $self->error("You must specify a user to login as.");
-            die $self->{'error'};                #exit eval
+        $self->{'remote_server'} = $self->{'ip'} || $self->{'host'};
+        $self->_validate_connection_settings();
+        if ( $self->{'operating_mode'} eq 'session' ) {
+            $self->_establish_session($service) if !( $self->{'security_tokens'}->{$service} && $self->{'cookie_jars'}->{$service} );
+            $self->{'ua'}->cookie_jar( $self->{'cookie_jars'}->{$service} );
         }
-        my $remote_server = $self->{'ip'}   || $self->{'host'};
-        my $server_name   = $self->{'host'} || $self->{'ip'};
-        if ( !$remote_server ) {
-            $self->error("You must set a host to connect to. (missing 'host' and 'ip' parameter)");
-            die $self->{'error'};                #exit eval
-        }
-        $server_name =~ s/\s*//g;
-        my $keepalive         = $self->{'keepalive'} ? 1 : 0;
-        my $connection_string = $remote_server . ':' . $port;
-        my $attempts          = 0;
-        my $hassigpipe;
+
+        my $remote_server    = $self->{'remote_server'};
+        my $attempts         = 0;
         my $finished_request = 0;
-        my $ssl_verify_mode  = $self->{'ssl_verify_mode'};
+        my $hassigpipe;
 
         local $SIG{'ALRM'} = sub {
             $self->error('Connection Timed Out');
             die $self->{'error'};
         };
+
         local $SIG{'PIPE'} = sub { $hassigpipe = 1; };
         $orig_alarm = alarm($timeout);
-        my @connection_args = ( PeerHost => $remote_server, PeerPort => $port, SSL_verify_mode => $ssl_verify_mode );
+
+        $formdata = $self->format_http_query($formdata) if ref $formdata;
+
+        my $scheme = $self->{'usessl'} ? "https" : "http";
+        my $url = "$scheme://$remote_server:$port";
+        if ( $self->{'operating_mode'} eq 'session' ) {
+            my $security_token = $self->{'security_tokens'}->{$service};
+            $url .= '/' . $self->{'security_tokens'}->{$service} . $uri;
+        }
+        else {
+            $url .= $uri;
+        }
+
+        my $content;
+        if ( $method eq 'POST' || $method eq 'PUT' ) {
+            $content = $formdata;
+        }
+        else {
+            $url .= "?$formdata";
+        }
+        $self->debug("URL: $url") if $self->{'debug'};
+
+        if ( !ref $headers ) {
+            my @lines = split /\r\n/, $headers;
+            $headers = {};
+            foreach my $line (@lines) {
+                last unless length $line;
+                my ( $key, $value ) = split /:\s*/, $line, 2;
+                next unless length $key;
+                $headers->{$key} ||= [];
+                push @{ $headers->{$key} }, $value;
+            }
+        }
+
+        my $options = {
+            headers => $headers,
+        };
+        $options->{'content'} = $content if defined $content;
+        my $ua = $self->{'ua'};
         while ( ++$attempts < 3 ) {
             $hassigpipe = 0;
-            if ( !ref $whm_sock || !$whm_sock_host || $whm_sock_host ne $connection_string ) {
-                close($whm_sock) if ref $whm_sock;
-                $whm_sock =
-                  $self->{'usessl'}
-                  ? IO::Socket::SSL->new(@connection_args)
-                  : IO::Socket::INET->new($connection_string);
-                if ( !$whm_sock ) {
-                    undef $whm_sock;
-                    $self->error("Could not connect to $connection_string: $!");
-                    die $self->{'error'};    #exit eval
-                }
-                else {
-                    $whm_sock_host = $connection_string;
-                }
-            }
-            my $connection_type = ( $keepalive ? 'Keep-Alive' : 'Close' );
-
-            $headers = $self->format_http_headers($headers) if ref $headers;
-            if ( $method eq 'POST' || $method eq 'PUT' ) {
-                $formdata = $self->format_http_query($formdata) if ref $formdata;
-                print {$whm_sock} "$method $uri HTTP/1.1\r\nHost: $server_name\r\nConnection: $connection_type\r\n" . 'Content-Length: ' . length($formdata) . "\r\n" . "User-Agent: " . 'cPanel::PublicAPI (perl) ' . $VERSION . "\r\n" . "Authorization: $authstr\r\n$headers\r\n" . $formdata;
-                $self->debug( ( "$method $uri HTTP/1.1\r\nHost: $server_name\r\nConnection: $connection_type\r\n" . 'Content-Length: ' . length($formdata) . "\r\n" . "User-Agent: " . 'cPanel::PublicAPI (perl) ' . $VERSION . "\r\n" . "Authorization: $authstr\r\n$headers\r\n" . $formdata ) ) if $self->{'debug'};
-            }
-            else {
-                print {$whm_sock} "$method "
-                  . ( $uri . '?' . ( ref $formdata ? join( '&', map { $CFG{'uri_encoder_func'}->($_) . '=' . $CFG{'uri_encoder_func'}->( $formdata->{$_} ) } keys %{$formdata} ) : $formdata ) )
-                  . " HTTP/1.1\r\nHost: $server_name\r\nConnection: $connection_type\r\n"
-                  . "User-Agent: "
-                  . 'cPanel::PublicAPI (perl) '
-                  . $VERSION . "\r\n"
-                  . "Authorization: $authstr\r\n$headers\r\n";
-
-                $self->debug( "$method " . ( $uri . '?' . ( ref $formdata ? join( '&', map { $CFG{'uri_encoder_func'}->($_) . '=' . $CFG{'uri_encoder_func'}->( $formdata->{$_} ) } keys %{$formdata} ) : $formdata ) ) . " HTTP/1.1\r\nHost: $server_name\r\nConnection: $connection_type\r\n" . "User-Agent: " . 'cPanel::PublicAPI (perl) ' . $VERSION . "\r\n" . "Authorization: $authstr\r\n$headers\r\n" )
-                  if $self->{'debug'};
-
+            my $response = $ua->request( $method, $url, $options );
+            if ( $response->{'status'} == 599 ) {
+                $self->error("Could not connect to $url: $response->{'content'}");
+                die $self->{'error'};    #exit eval
             }
 
-            if ($hassigpipe) { undef $whm_sock_host; next; }    # http spec says to reconnect
-
-            my $inheaders   = 1;
-            my $httpheader  = readline($whm_sock);
-            my $http_status = $httpheader;
-
-            $self->debug("HTTP STATUS[$http_status]") if $self->{'debug'};
-
-            if ( $hassigpipe || !$httpheader ) { undef $whm_sock_host; next; }    # http spec says to reconnect
-
+            if ($hassigpipe) { next; }    # http spec says to reconnect
             my %HEADERS;
-            {
-                local $/ = ( $httpheader =~ tr/\r// ) ? "\r\n\r\n" : "\n\n";
-                %HEADERS = map { ( lc $_->[0], substr( $_->[1], 0, 8190 ) ) }     # lc the header and truncate the value to 8190 bytes
-                  map { [ ( split( /:\s*/, $_, 2 ) )[ 0, 1 ] ] }                  # split header into name, value - and place into an arrayref for the next map to alter
-                  split( /\r?\n/, readline($whm_sock) );                          # split each header
-            }
             if ( $self->{'debug'} ) {
+                %HEADERS = %{ $response->{'headers'} };
                 foreach my $header ( keys %HEADERS ) {
                     $self->debug("HEADER[$header]=[$HEADERS{$header}]");
                 }
-            }
-            {
                 if ( exists $HEADERS{'transfer-encoding'} && $HEADERS{'transfer-encoding'} =~ /chunked/i ) {
-                    $self->debug("READ TYPE=chunked") if $self->{'debug'};
-                    local $/ = "\r\n";
-                    my ( $size_to_read, $read_size, $read_buffer, $bytes_read );
-                  CHUNKED_READ:
-                    while (1) {
-                        chomp( $read_size = readline($whm_sock) );
-                        $size_to_read = hex($read_size) + 2;
-
-                        #if ( $self->{'debug'} ) {
-                        #    $self->debug("CHUNKED read_size=$read_size")               ;
-                        #    $self->debug("REQUESTING_READ size_to_read=$size_to_read") ;
-                        #}
-                        while ($size_to_read) {
-                            if ( !defined($read_size) || !( $bytes_read = read( $whm_sock, $read_buffer, $size_to_read ) ) ) {
-
-                                #$self->debug("FINSIHED READING CHUNKED DATA [$bytes_read]") if $self->{'debug'};
-                                last CHUNKED_READ;
-                            }
-                            $size_to_read -= $bytes_read;
-
-                            #$self->debug("BYTES READ = [$bytes_read] (remaining in chunk $size_to_read)") if $self->{'debug'};
-                            $page .=
-                                $size_to_read == 0 ? substr( $read_buffer, 0, -2 )
-                              : $size_to_read == 1 ? substr( $read_buffer, 0, -1 )
-                              :                      $read_buffer;
-                        }
-                        last if hex($read_size) == 0;
-                    }
-                    $keepalive = 0 if exists $HEADERS{'connection'} && $HEADERS{'connection'} =~ /close/i;
+                    $self->debug("READ TYPE=chunked");
                 }
                 elsif ( defined $HEADERS{'content-length'} ) {
-                    $self->debug("READ TYPE=content-length") if $self->{'debug'};
-                    my $bytes_to_read = int $HEADERS{'content-length'};
-                    my $bytes_read = $bytes_to_read ? read( $whm_sock, $page, $bytes_to_read ) : 0;
-                    if ( ( $bytes_to_read -= $bytes_read ) > 0 ) {
-                        while ( $bytes_read = read( $whm_sock, $page, $bytes_to_read, length $page ) ) {
-                            last if !$bytes_read || !( $bytes_to_read -= $bytes_read );
-                        }
-                    }
-                    $self->debug("READ TYPE=content-length: Failed to read response from server.  The connection was dropped with $bytes_to_read bytes left: $!\n") if $self->{'debug'} && $bytes_to_read;
-                    $keepalive = 0 if exists $HEADERS{'connection'} && $HEADERS{'connection'} =~ /close/i;
+                    $self->debug("READ TYPE=content-length");
                 }
                 else {
-                    $self->debug("READ TYPE=close") if $self->{'debug'};
-                    $keepalive = 0;
-                    local $/;
-                    $page = readline($whm_sock);    # read until we close ths socket
+                    $self->debug("READ TYPE=close");
                 }
             }
 
-            if ( $http_status !~ m/^HTTP\/1\.(0|1) 2/ ) {
-                $keepalive = 0;
-                $self->error("Server Error from $remote_server: $http_status");
+            if ( !$response->{'success'} ) {
+                $self->error("Server Error from $remote_server: $response->{'status'} $response->{'reason'}");
             }
 
-            if ( !$keepalive ) {
-                close($whm_sock);
-                undef $whm_sock_host;
-            }
+            $page = $response->{'content'};
 
             $finished_request = 1;
             last;
@@ -438,6 +344,101 @@ sub api_request {
     return ( $self->{'error'} ? 0 : 1, $self->{'error'}, \$page );
 }
 
+sub establish_tfa_session {
+    my ( $self, $service, $tfa_token ) = @_;
+    if ( $self->{'operating_mode'} ne 'session' ) {
+        $self->error("2FA-authenticated sessions are not supported when using accesshash keys or API tokens");
+        die $self->{'error'};
+    }
+    if ( !( $service && $tfa_token ) ) {
+        $self->error("You must specify the service name, and the 2FA token in order to establish a 2FA-authenticated session");
+        die $self->{'error'};
+    }
+
+    undef $self->{'cookie_jars'}->{$service};
+    undef $self->{'security_tokens'}->{$service};
+    return $self->_establish_session( $service, $tfa_token );
+}
+
+sub _validate_connection_settings {
+    my $self = shift;
+
+    if ( !$self->{'user'} ) {
+        $self->error("You must specify a user to login as.");
+        die $self->{'error'};
+    }
+
+    if ( !$self->{'remote_server'} ) {
+        $self->error("You must set a host to connect to. (missing 'host' and 'ip' parameter)");
+        die $self->{'error'};
+    }
+}
+
+sub _update_operating_mode {
+    my $self = shift;
+
+    if ( exists $self->{'accesshash'} ) {
+        $self->{'accesshash'} =~ s/[\r\n]//g;
+        $self->{'ua'}->default_headers( { 'Authorization' => 'WHM ' . $self->{'user'} . ':' . $self->{'accesshash'} } );
+        $self->{'operating_mode'} = 'accesshash';
+    }
+    elsif ( exists $self->{'pass'} ) {
+        $self->{'operating_mode'} = 'session';
+
+        # This is called whenever the pass or accesshash is changed,
+        # so we reset the cookie jars, and tokens on such changes
+        $self->{'cookie_jars'}     = { map { $_ => undef } keys %PORT_DB };
+        $self->{'security_tokens'} = { map { $_ => undef } keys %PORT_DB };
+    }
+    else {
+        $self->error('You must specify an accesshash, API token, or password');
+        die $self->{'error'};
+    }
+}
+
+sub _establish_session {
+    my ( $self, $service, $tfa_token ) = @_;
+
+    return if $self->{'operating_mode'} ne 'session';
+    return if $self->{'security_tokens'}->{$service} && $self->{'cookie_jars'}->{$service};
+
+    $self->{'cookie_jars'}->{$service} = HTTP::CookieJar->new();
+    $self->{'ua'}->cookie_jar( $self->{'cookie_jars'}->{$service} );
+
+    my $port   = $self->_determine_port_for_service($service);
+    my $scheme = $self->{'usessl'} ? "https" : "http";
+    my $url    = "$scheme://$self->{'remote_server'}:$port/login";
+    my $resp   = $self->{'ua'}->post_form(
+        $url,
+        {
+            'user' => $self->{'user'},
+            'pass' => $self->{'pass'},
+            ( $tfa_token ? ( 'tfa_token' => $tfa_token ) : () ),
+        },
+    );
+
+    if ( my $security_token = ( split /\//, $resp->{'headers'}->{'location'} )[1] ) {
+        $self->{'security_tokens'}->{$service} = $security_token;
+        return 1;
+    }
+
+    $self->error("Failed to establish session and parse security token: $resp->{'status'} $resp->{'reason'}");
+    die $self->{'error'};
+}
+
+sub _determine_port_for_service {
+    my ( $self, $service ) = @_;
+
+    my $port;
+    if ( $self->{'usessl'} ) {
+        $port = $service =~ /^\d+$/ ? $service : $PORT_DB{$service}{'ssl'};
+    }
+    else {
+        $port = $service =~ /^\d+$/ ? $service : $PORT_DB{$service}{'plaintext'};
+    }
+    return $port;
+}
+
 sub cpanel_api1_request {
     my ( $self, $service, $cfg, $formdata, $format ) = @_;
 
@@ -460,22 +461,14 @@ sub cpanel_api1_request {
     $formdata->{ 'cpanel_' . $query_format . 'api_apiversion' } = 1;
 
     my ( $status, $statusmsg, $data ) = $self->api_request( $service, '/' . $query_format . '-api/cpanel', ( ( scalar keys %$formdata < 10 && _total_form_length( $formdata, 1024 ) < 1024 ) ? 'GET' : 'POST' ), $formdata );
-    if ( defined $format && ( $format eq 'json' || $format eq 'xml' ) ) {
-        return $$data;
-    }
-    else {
-        my $parsed_data;
-        eval { $parsed_data = $CFG{'serializer_can_deref'} ? $CFG{'api_serializer_obj'}->($data) : $CFG{'api_serializer_obj'}->($$data); };
-        if ( !ref $parsed_data ) {
-            $self->error("There was an issue with parsing the following response from cPanel or WHM:\n$$data");
-            die $self->{'error'};
+
+    return $self->_parse_returndata(
+        {
+            'caller' => 'cpanel_api1',
+            'data'   => $data,
+            'format' => $format,
         }
-        if ( exists $parsed_data->{'event'}->{'reason'} && $parsed_data->{'event'}->{'reason'} =~ /failed: Undefined subroutine/ ) {
-            $self->error( "cPanel::PublicAPI::cpanel_api1_request was called with the invalid API1 call of: " . $parsed_data->{'module'} . '::' . $parsed_data->{'func'} );
-            return;
-        }
-        return $parsed_data;
-    }
+    );
 }
 
 sub cpanel_api2_request {
@@ -496,24 +489,83 @@ sub cpanel_api2_request {
     $formdata->{ 'cpanel_' . $query_format . 'api_apiversion' } = 2;
     my ( $status, $statusmsg, $data ) = $self->api_request( $service, '/' . $query_format . '-api/cpanel', ( ( scalar keys %$formdata < 10 && _total_form_length( $formdata, 1024 ) < 1024 ) ? 'GET' : 'POST' ), $formdata );
 
-    if ( defined $format && ( $format eq 'json' || $format eq 'xml' ) ) {
-        return $$data;
+    return $self->_parse_returndata(
+        {
+            'caller' => 'cpanel_api2',
+            'data'   => $data,
+            'format' => $format,
+        }
+    );
+}
+
+sub _parse_returndata {
+    my ( $self, $opts_hr ) = @_;
+
+    if ( $self->{'error'} ) {
+        die $self->{'error'};
+    }
+    elsif ( ${ $opts_hr->{'data'} } =~ m/tfa_login_form/ ) {
+        $self->error("Two-Factor Authentication enabled on the account. Establish a session with the security token, or disable 2FA on the account");
+        die $self->{'error'};
+    }
+
+    if ( defined $opts_hr->{'format'} && ( $opts_hr->{'format'} eq 'json' || $opts_hr->{'format'} eq 'xml' ) ) {
+        return ${ $opts_hr->{'data'} };
     }
     else {
         my $parsed_data;
-        eval { $parsed_data = $CFG{'serializer_can_deref'} ? $CFG{'api_serializer_obj'}->($data) : $CFG{'api_serializer_obj'}->($$data); };
+        eval { $parsed_data = $CFG{'serializer_can_deref'} ? $CFG{'api_serializer_obj'}->( $opts_hr->{'data'} ) : $CFG{'api_serializer_obj'}->( ${ $opts_hr->{'data'} } ); };
         if ( !ref $parsed_data ) {
-            $self->error("There was an issue with parsing the following response from cPanel or WHM:\n$$data");
+            $self->error("There was an issue with parsing the following response from cPanel or WHM: [data=[${$opts_hr->{'data'}}]]");
             die $self->{'error'};
         }
-        if ( exists $parsed_data->{'cpanelresult'}->{'error'} && $parsed_data->{'cpanelresult'}->{'error'} =~ /Could not find function/ ) {    # xml-api v1 version
-            $self->error( "cPanel::PublicAPI::cpanel_api2_request was called with the invalid API2 call of: " . $parsed_data->{'cpanelresult'}->{'module'} . '::' . $parsed_data->{'cpanelresult'}->{'func'} );
-            return;
-        }
-        return $parsed_data;
+
+        my $error_check_dt = {
+            'whm_api'     => \&_check_whm_api_errors,
+            'cpanel_api1' => \&_check_cpanel_api1_errors,
+            'cpanel_api2' => \&_check_cpanel_api2_errors,
+        };
+        return $error_check_dt->{ $opts_hr->{'caller'} }->( $self, $opts_hr->{'call'}, $parsed_data );
     }
 }
 
+sub _check_whm_api_errors {
+    my ( $self, $call, $parsed_data ) = @_;
+
+    if (
+        ( exists $parsed_data->{'error'} && $parsed_data->{'error'} =~ /Unknown App Requested/ ) ||    # xml-api v0 version
+        ( exists $parsed_data->{'metadata'}->{'reason'} && $parsed_data->{'metadata'}->{'reason'} =~ /Unknown app\s+(?:\(.+\))?\s+requested/ )    # xml-api v1 version
+      ) {
+        $self->error("cPanel::PublicAPI::whm_api was called with the invalid API call of: $call.");
+        return;
+    }
+    return $parsed_data;
+}
+
+sub _check_cpanel_api1_errors {
+    my ( $self, undef, $parsed_data ) = @_;
+    if (
+        exists $parsed_data->{'event'}->{'reason'} && (
+            $parsed_data->{'event'}->{'reason'} =~ /failed: Undefined subroutine/ ||                                                              # pre-11.44 error message
+            $parsed_data->{'event'}->{'reason'} =~ m/failed: Can\'t use string/                                                                   # 11.44+ error message
+        )
+      ) {
+        $self->error( "cPanel::PublicAPI::cpanel_api1_request was called with the invalid API1 call of: " . $parsed_data->{'module'} . '::' . $parsed_data->{'func'} );
+        return;
+    }
+    return $parsed_data;
+}
+
+sub _check_cpanel_api2_errors {
+    my ( $self, undef, $parsed_data ) = @_;
+
+    if ( exists $parsed_data->{'cpanelresult'}->{'error'} && $parsed_data->{'cpanelresult'}->{'error'} =~ /Could not find function/ ) {           # xml-api v1 version
+        $self->error( "cPanel::PublicAPI::cpanel_api2_request was called with the invalid API2 call of: " . $parsed_data->{'cpanelresult'}->{'module'} . '::' . $parsed_data->{'cpanelresult'}->{'func'} );
+        return;
+    }
+    return $parsed_data;
+}
+
 sub _total_form_length {
     my $data = shift;
     my $max  = shift;
@@ -610,3 +662,4 @@ sub format_http_query {
     }
     return $formdata;
 }
+
